@@ -1,11 +1,10 @@
 use crate::{
-    backend::{error::ContextError, native_gpu_future},
-    AdapterInfo, BindGroupDescriptor, BindGroupLayoutDescriptor, BindingResource, BufferBinding,
-    CommandEncoderDescriptor, ComputePassDescriptor, ComputePipelineDescriptor,
-    DownlevelProperties, Features, Label, Limits, LoadOp, MapMode, Operations,
-    PipelineLayoutDescriptor, RenderBundleEncoderDescriptor, RenderPipelineDescriptor,
-    SamplerDescriptor, ShaderModuleDescriptor, ShaderSource, SwapChainStatus, TextureDescriptor,
-    TextureFormat, TextureViewDescriptor,
+    backend::native_gpu_future, AdapterInfo, BindGroupDescriptor, BindGroupLayoutDescriptor,
+    BindingResource, BufferBinding, CommandEncoderDescriptor, ComputePassDescriptor,
+    ComputePipelineDescriptor, DownlevelCapabilities, Features, Label, Limits, LoadOp, MapMode,
+    Operations, PipelineLayoutDescriptor, RenderBundleEncoderDescriptor, RenderPipelineDescriptor,
+    SamplerDescriptor, ShaderModuleDescriptor, ShaderModuleDescriptorSpirV, ShaderSource,
+    SurfaceStatus, TextureDescriptor, TextureFormat, TextureViewDescriptor,
 };
 
 use arrayvec::ArrayVec;
@@ -39,17 +38,91 @@ impl fmt::Debug for Context {
 }
 
 impl Context {
+    pub unsafe fn from_hal_instance<A: wgc::hub::HalApi>(hal_instance: A::Instance) -> Self {
+        Self(wgc::hub::Global::from_hal_instance::<A>(
+            "wgpu",
+            wgc::hub::IdentityManagerFactory,
+            hal_instance,
+        ))
+    }
+
     pub(crate) fn global(&self) -> &wgc::hub::Global<wgc::hub::IdentityManagerFactory> {
         &self.0
     }
 
-    pub fn enumerate_adapters(&self, backends: wgt::BackendBit) -> Vec<wgc::id::AdapterId> {
+    pub fn enumerate_adapters(&self, backends: wgt::Backends) -> Vec<wgc::id::AdapterId> {
         self.0
             .enumerate_adapters(wgc::instance::AdapterInputs::Mask(backends, |_| {
                 PhantomData
             }))
     }
 
+    pub unsafe fn create_adapter_from_hal<A: wgc::hub::HalApi>(
+        &self,
+        hal_adapter: hal::ExposedAdapter<A>,
+    ) -> wgc::id::AdapterId {
+        self.0.create_adapter_from_hal(hal_adapter, PhantomData)
+    }
+
+    pub unsafe fn create_device_from_hal<A: wgc::hub::HalApi>(
+        &self,
+        adapter: &wgc::id::AdapterId,
+        hal_device: hal::OpenDevice<A>,
+        desc: &crate::DeviceDescriptor,
+        trace_dir: Option<&std::path::Path>,
+    ) -> Result<(Device, wgc::id::QueueId), crate::RequestDeviceError> {
+        let global = &self.0;
+        let (device_id, error) = global.create_device_from_hal(
+            *adapter,
+            hal_device,
+            &desc.map_label(|l| l.map(Borrowed)),
+            trace_dir,
+            PhantomData,
+        );
+        if let Some(err) = error {
+            self.handle_error_fatal(err, "Adapter::create_device_from_hal");
+        }
+        let device = Device {
+            id: device_id,
+            error_sink: Arc::new(Mutex::new(ErrorSinkRaw::new())),
+            features: desc.features,
+        };
+        Ok((device, device_id))
+    }
+
+    pub unsafe fn create_texture_from_hal<A: wgc::hub::HalApi>(
+        &self,
+        hal_texture: A::Texture,
+        device: &Device,
+        desc: &TextureDescriptor,
+    ) -> Texture {
+        let global = &self.0;
+        let (id, error) = global.create_texture_from_hal::<A>(
+            hal_texture,
+            device.id,
+            &desc.map_label(|l| l.map(Borrowed)),
+            PhantomData,
+        );
+        if let Some(cause) = error {
+            self.handle_error(
+                &device.error_sink,
+                cause,
+                LABEL,
+                desc.label,
+                "Device::create_texture_from_hal",
+            );
+        }
+        Texture {
+            id,
+            error_sink: Arc::clone(&device.error_sink),
+        }
+    }
+
+    pub fn generate_report(&self) -> wgc::hub::GlobalReport {
+        self.0.generate_report()
+    }
+
+    /*TODO: raw surface
     #[cfg(any(target_os = "ios", target_os = "macos"))]
     pub unsafe fn create_surface_from_core_animation_layer(
         self: &Arc<Self>,
@@ -60,7 +133,7 @@ impl Context {
             context: Arc::clone(self),
             id,
         }
-    }
+    }*/
 
     fn handle_error(
         &self,
@@ -70,7 +143,7 @@ impl Context {
         label: Label,
         string: &'static str,
     ) {
-        let error = ContextError {
+        let error = wgc::error::ContextError {
             string,
             cause: Box::new(cause),
             label: label.unwrap_or_default().to_string(),
@@ -112,6 +185,25 @@ impl Context {
     ) -> ! {
         panic!("Error in {}: {}", string, cause);
     }
+
+    fn format_error(&self, err: &(impl Error + 'static)) -> String {
+        let global = self.global();
+        let mut err_descs = vec![];
+
+        let mut err_str = String::new();
+        wgc::error::format_pretty_any(&mut err_str, global, err);
+        err_descs.push(err_str);
+
+        let mut source_opt = err.source();
+        while let Some(source) = source_opt {
+            let mut source_str = String::new();
+            wgc::error::format_pretty_any(&mut source_str, global, source);
+            err_descs.push(source_str);
+            source_opt = source.source();
+        }
+
+        format!("Validation Error\n\nCaused by:\n{}", err_descs.join(""))
+    }
 }
 
 mod pass_impl {
@@ -234,7 +326,7 @@ mod pass_impl {
         ) {
             wgpu_render_pass_set_vertex_buffer(self, slot, buffer.id, offset, size)
         }
-        fn set_push_constants(&mut self, stages: wgt::ShaderStage, offset: u32, data: &[u8]) {
+        fn set_push_constants(&mut self, stages: wgt::ShaderStages, offset: u32, data: &[u8]) {
             unsafe {
                 wgpu_render_pass_set_push_constants(
                     self,
@@ -445,7 +537,7 @@ mod pass_impl {
             wgpu_render_bundle_set_vertex_buffer(self, slot, buffer.id, offset, size)
         }
 
-        fn set_push_constants(&mut self, stages: wgt::ShaderStage, offset: u32, data: &[u8]) {
+        fn set_push_constants(&mut self, stages: wgt::ShaderStages, offset: u32, data: &[u8]) {
             unsafe {
                 wgpu_render_bundle_set_push_constants(
                     self,
@@ -540,6 +632,7 @@ fn map_texture_copy_view(view: crate::ImageCopyTexture) -> wgc::command::ImageCo
         texture: view.texture.id.id,
         mip_level: view.mip_level,
         origin: view.origin,
+        aspect: view.aspect,
     }
 }
 
@@ -555,7 +648,7 @@ fn map_pass_channel<V: Copy + Default>(
             store_op: if store {
                 wgc::command::StoreOp::Store
             } else {
-                wgc::command::StoreOp::Clear
+                wgc::command::StoreOp::Discard
             },
             clear_value,
             read_only: false,
@@ -568,7 +661,7 @@ fn map_pass_channel<V: Copy + Default>(
             store_op: if store {
                 wgc::command::StoreOp::Store
             } else {
-                wgc::command::StoreOp::Clear
+                wgc::command::StoreOp::Discard
             },
             clear_value: V::default(),
             read_only: false,
@@ -583,7 +676,15 @@ fn map_pass_channel<V: Copy + Default>(
 }
 
 #[derive(Debug)]
-pub(crate) struct Device {
+pub struct Surface {
+    id: wgc::id::SurfaceId,
+    /// Configured device is needed to know which backend
+    /// code to execute when acquiring a new frame.
+    configured_device: Mutex<Option<wgc::id::DeviceId>>,
+}
+
+#[derive(Debug)]
+pub struct Device {
     id: wgc::id::DeviceId,
     error_sink: ErrorSink,
     features: Features,
@@ -596,7 +697,7 @@ pub(crate) struct Buffer {
 }
 
 #[derive(Debug)]
-pub(crate) struct Texture {
+pub struct Texture {
     id: wgc::id::TextureId,
     error_sink: ErrorSink,
 }
@@ -629,18 +730,18 @@ impl crate::Context for Context {
     type CommandBufferId = wgc::id::CommandBufferId;
     type RenderBundleEncoderId = wgc::command::RenderBundleEncoder;
     type RenderBundleId = wgc::id::RenderBundleId;
-    type SurfaceId = wgc::id::SurfaceId;
-    type SwapChainId = wgc::id::SwapChainId;
+    type SurfaceId = Surface;
 
-    type SwapChainOutputDetail = SwapChainOutputDetail;
+    type SurfaceOutputDetail = SurfaceOutputDetail;
 
     type RequestAdapterFuture = Ready<Option<Self::AdapterId>>;
     #[allow(clippy::type_complexity)]
     type RequestDeviceFuture =
         Ready<Result<(Self::DeviceId, Self::QueueId), crate::RequestDeviceError>>;
     type MapAsyncFuture = native_gpu_future::GpuFuture<Result<(), crate::BufferAsyncError>>;
+    type OnSubmittedWorkDoneFuture = native_gpu_future::GpuFuture<()>;
 
-    fn init(backends: wgt::BackendBit) -> Self {
+    fn init(backends: wgt::Backends) -> Self {
         Self(wgc::hub::Global::new(
             "wgpu",
             wgc::hub::IdentityManagerFactory,
@@ -652,7 +753,10 @@ impl crate::Context for Context {
         &self,
         handle: &impl raw_window_handle::HasRawWindowHandle,
     ) -> Self::SurfaceId {
-        self.0.instance_create_surface(handle, PhantomData)
+        Surface {
+            id: self.0.instance_create_surface(handle, PhantomData),
+            configured_device: Mutex::new(None),
+        }
     }
 
     fn instance_request_adapter(
@@ -662,9 +766,9 @@ impl crate::Context for Context {
         let id = self.0.request_adapter(
             &wgc::instance::RequestAdapterOptions {
                 power_preference: options.power_preference,
-                compatible_surface: options.compatible_surface.map(|surface| surface.id),
+                compatible_surface: options.compatible_surface.map(|surface| surface.id.id),
             },
-            wgc::instance::AdapterInputs::Mask(wgt::BackendBit::all(), |_| PhantomData),
+            wgc::instance::AdapterInputs::Mask(wgt::Backends::all(), |_| PhantomData),
         );
         ready(id.ok())
     }
@@ -691,7 +795,8 @@ impl crate::Context for Context {
             PhantomData
         ));
         if let Some(err) = error {
-            self.handle_error_fatal(err, "Adapter::request_device");
+            log::error!("Error in Adapter::request_device: {}", err);
+            return ready(Err(crate::RequestDeviceError));
         }
         let device = Device {
             id: device_id,
@@ -701,17 +806,16 @@ impl crate::Context for Context {
         ready(Ok((device, device_id)))
     }
 
-    fn adapter_get_swap_chain_preferred_format(
+    fn adapter_is_surface_supported(
         &self,
         adapter: &Self::AdapterId,
         surface: &Self::SurfaceId,
-    ) -> Option<TextureFormat> {
+    ) -> bool {
         let global = &self.0;
-        match wgc::gfx_select!(adapter => global.adapter_get_swap_chain_preferred_format(*adapter, *surface))
+        match wgc::gfx_select!(adapter => global.adapter_is_surface_supported(*adapter, surface.id))
         {
-            Ok(swap_chain_preferred_format) => Some(swap_chain_preferred_format),
-            Err(wgc::instance::GetSwapChainPreferredFormatError::UnsupportedQueueFamily) => None,
-            Err(err) => self.handle_error_fatal(err, "Adapter::get_swap_chain_preferred_format"),
+            Ok(result) => result,
+            Err(err) => self.handle_error_fatal(err, "Adapter::is_surface_supported"),
         }
     }
 
@@ -731,7 +835,7 @@ impl crate::Context for Context {
         }
     }
 
-    fn adapter_downlevel_properties(&self, adapter: &Self::AdapterId) -> DownlevelProperties {
+    fn adapter_downlevel_properties(&self, adapter: &Self::AdapterId) -> DownlevelCapabilities {
         let global = &self.0;
         match wgc::gfx_select!(*adapter => global.adapter_downlevel_properties(*adapter)) {
             Ok(downlevel) => downlevel,
@@ -760,6 +864,74 @@ impl crate::Context for Context {
         }
     }
 
+    fn surface_get_preferred_format(
+        &self,
+        surface: &Self::SurfaceId,
+        adapter: &Self::AdapterId,
+    ) -> Option<TextureFormat> {
+        let global = &self.0;
+        match wgc::gfx_select!(adapter => global.surface_get_preferred_format(surface.id, *adapter))
+        {
+            Ok(format) => Some(format),
+            Err(wgc::instance::GetSurfacePreferredFormatError::UnsupportedQueueFamily) => None,
+            Err(err) => self.handle_error_fatal(err, "Surface::get_preferred_format"),
+        }
+    }
+
+    fn surface_configure(
+        &self,
+        surface: &Self::SurfaceId,
+        device: &Self::DeviceId,
+        config: &wgt::SurfaceConfiguration,
+    ) {
+        let global = &self.0;
+        let error =
+            wgc::gfx_select!(device.id => global.surface_configure(surface.id, device.id, config));
+        if let Some(e) = error {
+            self.handle_error_fatal(e, "Surface::configure");
+        } else {
+            *surface.configured_device.lock() = Some(device.id);
+        }
+    }
+
+    fn surface_get_current_texture(
+        &self,
+        surface: &Self::SurfaceId,
+    ) -> (
+        Option<Self::TextureId>,
+        SurfaceStatus,
+        Self::SurfaceOutputDetail,
+    ) {
+        let global = &self.0;
+        let device_id = surface
+            .configured_device
+            .lock()
+            .expect("Surface was not configured?");
+        match wgc::gfx_select!(
+            device_id => global.surface_get_current_texture(surface.id, PhantomData)
+        ) {
+            Ok(wgc::present::SurfaceOutput { status, texture_id }) => (
+                texture_id.map(|id| Texture {
+                    id,
+                    error_sink: Arc::new(Mutex::new(ErrorSinkRaw::new())),
+                }),
+                status,
+                SurfaceOutputDetail {
+                    surface_id: surface.id,
+                },
+            ),
+            Err(err) => self.handle_error_fatal(err, "Surface::get_current_texture_view"),
+        }
+    }
+
+    fn surface_present(&self, texture: &Self::TextureId, detail: &Self::SurfaceOutputDetail) {
+        let global = &self.0;
+        match wgc::gfx_select!(texture.id => global.surface_present(detail.surface_id)) {
+            Ok(_status) => (),
+            Err(err) => self.handle_error_fatal(err, "Surface::present"),
+        }
+    }
+
     fn device_features(&self, device: &Self::DeviceId) -> Features {
         let global = &self.0;
         match wgc::gfx_select!(device.id => global.device_features(device.id)) {
@@ -776,7 +948,7 @@ impl crate::Context for Context {
         }
     }
 
-    fn device_downlevel_properties(&self, device: &Self::DeviceId) -> DownlevelProperties {
+    fn device_downlevel_properties(&self, device: &Self::DeviceId) -> DownlevelCapabilities {
         let global = &self.0;
         match wgc::gfx_select!(device.id => global.device_downlevel_properties(device.id)) {
             Ok(limits) => limits,
@@ -784,20 +956,6 @@ impl crate::Context for Context {
         }
     }
 
-    fn device_create_swap_chain(
-        &self,
-        device: &Self::DeviceId,
-        surface: &Self::SurfaceId,
-        desc: &wgt::SwapChainDescriptor,
-    ) -> Self::SwapChainId {
-        let global = &self.0;
-        let (sc, error) = wgc::gfx_select!(device.id => global.device_create_swap_chain(device.id, *surface, desc));
-        match error {
-            Some(e) => self.handle_error_fatal(e, "Device::create_swap_chain"),
-            None => sc,
-        }
-    }
-
     fn device_create_shader_module(
         &self,
         device: &Self::DeviceId,
@@ -806,10 +964,20 @@ impl crate::Context for Context {
         let global = &self.0;
         let descriptor = wgc::pipeline::ShaderModuleDescriptor {
             label: desc.label.map(Borrowed),
-            flags: desc.flags,
         };
         let source = match desc.source {
-            ShaderSource::SpirV(ref spv) => wgc::pipeline::ShaderModuleSource::SpirV(Borrowed(spv)),
+            #[cfg(feature = "spirv")]
+            ShaderSource::SpirV(ref spv) => {
+                // Parse the given shader code and store its representation.
+                let options = naga::front::spv::Options {
+                    adjust_coordinate_space: false, // we require NDC_Y_UP feature
+                    strict_capabilities: true,
+                    flow_graph_dump_prefix: None,
+                };
+                let parser = naga::front::spv::Parser::new(spv.iter().cloned(), &options);
+                let module = parser.parse().unwrap();
+                wgc::pipeline::ShaderModuleSource::Naga(module)
+            }
             ShaderSource::Wgsl(ref code) => wgc::pipeline::ShaderModuleSource::Wgsl(Borrowed(code)),
         };
         let (id, error) = wgc::gfx_select!(
@@ -827,6 +995,30 @@ impl crate::Context for Context {
         id
     }
 
+    unsafe fn device_create_shader_module_spirv(
+        &self,
+        device: &Self::DeviceId,
+        desc: &ShaderModuleDescriptorSpirV,
+    ) -> Self::ShaderModuleId {
+        let global = &self.0;
+        let descriptor = wgc::pipeline::ShaderModuleDescriptor {
+            label: desc.label.map(Borrowed),
+        };
+        let (id, error) = wgc::gfx_select!(
+            device.id => global.device_create_shader_module_spirv(device.id, &descriptor, Borrowed(&desc.source), PhantomData)
+        );
+        if let Some(cause) = error {
+            self.handle_error(
+                &device.error_sink,
+                cause,
+                LABEL,
+                desc.label,
+                "Device::create_shader_module_spirv",
+            );
+        }
+        id
+    }
+
     fn device_create_bind_group_layout(
         &self,
         device: &Self::DeviceId,
@@ -860,10 +1052,7 @@ impl crate::Context for Context {
         use wgc::binding_model as bm;
 
         let mut arrayed_texture_views = Vec::new();
-        if device
-            .features
-            .contains(Features::SAMPLED_TEXTURE_BINDING_ARRAY)
-        {
+        if device.features.contains(Features::TEXTURE_BINDING_ARRAY) {
             // gather all the array view IDs first
             for entry in desc.entries.iter() {
                 if let BindingResource::TextureViewArray(array) = entry.resource {
@@ -951,24 +1140,24 @@ impl crate::Context for Context {
         device: &Self::DeviceId,
         desc: &PipelineLayoutDescriptor,
     ) -> Self::PipelineLayoutId {
-        // Limit is always less or equal to wgc::MAX_BIND_GROUPS, so this is always right
+        // Limit is always less or equal to hal::MAX_BIND_GROUPS, so this is always right
         // Guards following ArrayVec
         assert!(
-            desc.bind_group_layouts.len() <= wgc::MAX_BIND_GROUPS,
+            desc.bind_group_layouts.len() <= hal::MAX_BIND_GROUPS,
             "Bind group layout count {} exceeds device bind group limit {}",
             desc.bind_group_layouts.len(),
-            wgc::MAX_BIND_GROUPS
+            hal::MAX_BIND_GROUPS
         );
 
         let temp_layouts = desc
             .bind_group_layouts
             .iter()
             .map(|bgl| bgl.id)
-            .collect::<ArrayVec<[_; wgc::MAX_BIND_GROUPS]>>();
+            .collect::<ArrayVec<_, { hal::MAX_BIND_GROUPS }>>();
         let descriptor = wgc::binding_model::PipelineLayoutDescriptor {
             label: desc.label.map(Borrowed),
             bind_group_layouts: Borrowed(&temp_layouts),
-            push_constant_ranges: Borrowed(&desc.push_constant_ranges),
+            push_constant_ranges: Borrowed(desc.push_constant_ranges),
         };
 
         let global = &self.0;
@@ -996,7 +1185,7 @@ impl crate::Context for Context {
     ) -> Self::RenderPipelineId {
         use wgc::pipeline as pipe;
 
-        let vertex_buffers: ArrayVec<[_; wgc::device::MAX_VERTEX_BUFFERS]> = desc
+        let vertex_buffers: ArrayVec<_, { hal::MAX_VERTEX_BUFFERS }> = desc
             .vertex
             .buffers
             .iter()
@@ -1011,7 +1200,7 @@ impl crate::Context for Context {
             Some(_) => None,
             None => Some(wgc::device::ImplicitPipelineIds {
                 root_id: PhantomData,
-                group_ids: &[PhantomData; wgc::MAX_BIND_GROUPS],
+                group_ids: &[PhantomData; hal::MAX_BIND_GROUPS],
             }),
         };
         let descriptor = pipe::RenderPipelineDescriptor {
@@ -1047,7 +1236,6 @@ impl crate::Context for Context {
             if let wgc::pipeline::CreateRenderPipelineError::Internal { stage, ref error } = cause {
                 log::warn!("Shader translation error for stage {:?}: {}", stage, error);
                 log::warn!("Please report it to https://github.com/gfx-rs/naga");
-                log::warn!("Try enabling `wgpu/cross` feature as a workaround.");
             }
             self.handle_error(
                 &device.error_sink,
@@ -1071,7 +1259,7 @@ impl crate::Context for Context {
             Some(_) => None,
             None => Some(wgc::device::ImplicitPipelineIds {
                 root_id: PhantomData,
-                group_ids: &[PhantomData; wgc::MAX_BIND_GROUPS],
+                group_ids: &[PhantomData; hal::MAX_BIND_GROUPS],
             }),
         };
         let descriptor = pipe::ComputePipelineDescriptor {
@@ -1094,11 +1282,10 @@ impl crate::Context for Context {
             if let wgc::pipeline::CreateComputePipelineError::Internal(ref error) = cause {
                 log::warn!(
                     "Shader translation error for stage {:?}: {}",
-                    wgt::ShaderStage::COMPUTE,
+                    wgt::ShaderStages::COMPUTE,
                     error
                 );
                 log::warn!("Please report it to https://github.com/gfx-rs/naga");
-                log::warn!("Try enabling `wgpu/cross` feature as a workaround.");
             }
             self.handle_error(
                 &device.error_sink,
@@ -1206,12 +1393,12 @@ impl crate::Context for Context {
     fn device_create_query_set(
         &self,
         device: &Self::DeviceId,
-        desc: &wgt::QuerySetDescriptor,
+        desc: &wgt::QuerySetDescriptor<Label>,
     ) -> Self::QuerySetId {
         let global = &self.0;
         let (id, error) = wgc::gfx_select!(device.id => global.device_create_query_set(
             device.id,
-            &desc,
+            &desc.map_label(|l| l.map(Borrowed)),
             PhantomData
         ));
         if let Some(cause) = error {
@@ -1255,7 +1442,7 @@ impl crate::Context for Context {
         let descriptor = wgc::command::RenderBundleEncoderDescriptor {
             label: desc.label.map(Borrowed),
             color_formats: Borrowed(desc.color_formats),
-            depth_stencil_format: desc.depth_stencil_format,
+            depth_stencil: desc.depth_stencil,
             sample_count: desc.sample_count,
         };
         match wgc::command::RenderBundleEncoder::new(&descriptor, device.id, None) {
@@ -1372,37 +1559,6 @@ impl crate::Context for Context {
         }
     }
 
-    fn swap_chain_get_current_texture_view(
-        &self,
-        swap_chain: &Self::SwapChainId,
-    ) -> (
-        Option<Self::TextureViewId>,
-        SwapChainStatus,
-        Self::SwapChainOutputDetail,
-    ) {
-        let global = &self.0;
-        match wgc::gfx_select!(
-            *swap_chain => global.swap_chain_get_current_texture_view(*swap_chain, PhantomData)
-        ) {
-            Ok(wgc::swap_chain::SwapChainOutput { status, view_id }) => (
-                view_id,
-                status,
-                SwapChainOutputDetail {
-                    swap_chain_id: *swap_chain,
-                },
-            ),
-            Err(err) => self.handle_error_fatal(err, "SwapChain::get_current_texture_view"),
-        }
-    }
-
-    fn swap_chain_present(&self, view: &Self::TextureViewId, detail: &Self::SwapChainOutputDetail) {
-        let global = &self.0;
-        match wgc::gfx_select!(*view => global.swap_chain_present(detail.swap_chain_id)) {
-            Ok(_status) => (),
-            Err(err) => self.handle_error_fatal(err, "SwapChain::present"),
-        }
-    }
-
     fn texture_create_view(
         &self,
         texture: &Self::TextureId,
@@ -1731,7 +1887,7 @@ impl crate::Context for Context {
                 resolve_target: ca.resolve_target.map(|rt| rt.id),
                 channel: map_pass_channel(Some(&ca.ops)),
             })
-            .collect::<ArrayVec<[_; wgc::device::MAX_COLOR_TARGETS]>>();
+            .collect::<ArrayVec<_, { hal::MAX_COLOR_TARGETS }>>();
 
         let depth_stencil = desc.depth_stencil_attachment.as_ref().map(|dsa| {
             wgc::command::RenderPassDepthStencilAttachment {
@@ -1818,7 +1974,7 @@ impl crate::Context for Context {
 
     fn command_encoder_insert_debug_marker(&self, encoder: &Self::CommandEncoderId, label: &str) {
         let global = &self.0;
-        if let Err(cause) = wgc::gfx_select!(encoder.id => global.command_encoder_insert_debug_marker(encoder.id, &label))
+        if let Err(cause) = wgc::gfx_select!(encoder.id => global.command_encoder_insert_debug_marker(encoder.id, label))
         {
             self.handle_error_nolabel(
                 &encoder.error_sink,
@@ -1829,7 +1985,7 @@ impl crate::Context for Context {
     }
     fn command_encoder_push_debug_group(&self, encoder: &Self::CommandEncoderId, label: &str) {
         let global = &self.0;
-        if let Err(cause) = wgc::gfx_select!(encoder.id => global.command_encoder_push_debug_group(encoder.id, &label))
+        if let Err(cause) = wgc::gfx_select!(encoder.id => global.command_encoder_push_debug_group(encoder.id, label))
         {
             self.handle_error_nolabel(
                 &encoder.error_sink,
@@ -1932,6 +2088,31 @@ impl crate::Context for Context {
         }
     }
 
+    fn queue_on_submitted_work_done(
+        &self,
+        queue: &Self::QueueId,
+    ) -> Self::OnSubmittedWorkDoneFuture {
+        let (future, completion) = native_gpu_future::new_gpu_future();
+
+        extern "C" fn submitted_work_done_future_wrapper(user_data: *mut u8) {
+            let completion =
+                unsafe { native_gpu_future::GpuFutureCompletion::from_raw(user_data as _) };
+            completion.complete(())
+        }
+
+        let closure = wgc::device::queue::SubmittedWorkDoneClosure {
+            callback: submitted_work_done_future_wrapper,
+            user_data: completion.into_raw() as _,
+        };
+
+        let global = &self.0;
+        let res = wgc::gfx_select!(queue => global.queue_on_submitted_work_done(*queue, closure));
+        if let Err(cause) = res {
+            self.handle_error_fatal(cause, "Queue::on_submitted_work_done");
+        }
+        future
+    }
+
     fn device_start_capture(&self, device: &Self::DeviceId) {
         let global = &self.0;
         wgc::gfx_select!(device.id => global.device_start_capture(device.id));
@@ -1944,8 +2125,8 @@ impl crate::Context for Context {
 }
 
 #[derive(Debug)]
-pub(crate) struct SwapChainOutputDetail {
-    swap_chain_id: wgc::id::SwapChainId,
+pub(crate) struct SurfaceOutputDetail {
+    surface_id: wgc::id::SurfaceId,
 }
 
 type ErrorSink = Arc<Mutex<ErrorSinkRaw>>;
