@@ -1,13 +1,13 @@
 //! Convenience macros
 
-/// Macro to produce an array of [VertexAttribute](crate::VertexAttribute).
+/// Macro to produce an array of [`VertexAttribute`](crate::VertexAttribute).
 ///
 /// Output has type: `[VertexAttribute; _]`. Usage is as follows:
 /// ```
 /// # use wgpu::vertex_attr_array;
 /// let attrs = vertex_attr_array![0 => Float32x2, 1 => Float32, 2 => Uint16x4];
 /// ```
-/// This example specifies a list of three [VertexAttribute](crate::VertexAttribute),
+/// This example specifies a list of three [`VertexAttribute`](crate::VertexAttribute),
 /// each with the given `shader_location` and `format`.
 /// Offsets are calculated automatically.
 #[macro_export]
@@ -45,6 +45,7 @@ fn test_vertex_attr_array() {
 ///
 /// It ensures the word alignment as well as the magic number.
 #[macro_export]
+#[cfg(feature = "spirv")]
 macro_rules! include_spirv {
     ($($token:tt)*) => {
         {
@@ -52,7 +53,24 @@ macro_rules! include_spirv {
             $crate::ShaderModuleDescriptor {
                 label: Some($($token)*),
                 source: $crate::util::make_spirv(include_bytes!($($token)*)),
-                flags: $crate::ShaderFlags::VALIDATION,
+            }
+        }
+    };
+}
+
+/// Macro to load raw SPIR-V data statically, for use with [`Features::SPIRV_SHADER_PASSTHROUGH`].
+///
+/// It ensures the word alignment as well as the magic number.
+///
+/// [`Features::SPIRV_SHADER_PASSTHROUGH`]: crate::Features::SPIRV_SHADER_PASSTHROUGH
+#[macro_export]
+macro_rules! include_spirv_raw {
+    ($($token:tt)*) => {
+        {
+            //log::info!("including '{}'", $($token)*);
+            $crate::ShaderModuleDescriptorSpirV {
+                label: Some($($token)*),
+                source: $crate::util::make_spirv_raw(include_bytes!($($token)*)),
             }
         }
     };
@@ -67,7 +85,6 @@ macro_rules! include_wgsl {
             $crate::ShaderModuleDescriptor {
                 label: Some($($token)*),
                 source: $crate::ShaderSource::Wgsl(include_str!($($token)*).into()),
-                flags: $crate::ShaderFlags::all(),
             }
         }
     };
